/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */ 
/**
 * @file bt_acl_com.h
 * 
 *\author Martin Hinz <btnode@hinz.ch>
 * 
 *\date 26.7.2004
 * 
 * @brief Implements communication over acl layer.
 * 
 * Every pkt can be given a service number. An application can get
 * the acl pkts on the basis of this service number.
 * 
 * pkt structs definition for better understanding :)
\verbatim
bt_stack:
                       -----------------------------------
struct bt_hci_pkt_acl: |type|       header       |payload|
                       -----------------------------------
                         1   HCI_ACL_DATA_PKT_HDR    1

bt_acl_com_stack:
                       ------------------------------------------------------------------------
bt_acl_com_pkt_t     : | 1  |HCI_ACL_DATA_PKT_HDR|BT_ACL_COM_HEADER_LEN|BT_MAX_ACL_COM_PAYLOAD|
(array of u_char)      ------------------------------------------------------------------------

                                                 ----------------------------------------------
bt_acl_com_t         :                           |  len  | service nr. |       payload        |
                                                 ----------------------------------------------
                                                     2          2       BT_MAX_ACL_COM_PAYLOAD
\endverbatim
 */


#ifndef _BT_ACL_COM_H_
#define _BT_ACL_COM_H_


#include <sys/types.h>
#include <bt/bt_defs.h>

/**
 * defines the maximal service nr 0 - (#MAX_SERVICE_NR-1) that can be given the
 * acl pkts
 */
#define MAX_SERVICE_NR 8

/**
 * acl pkt arriving error: invalid service number (due to communication problems)
 */
#define BT_ERR_INVALID_SERV_NR        1

/**
 * acl pkt arriving error: invalid connection handle (due to communication problems)
 */
#define BT_ERR_INVALID_HANDLE         2

/**
 * acl pkt arriving error: pkt fragmentation (due to wrong pkt type, use #HCI_PACKET_TYPE!)
 */
#define BT_ERR_FRAGMENTED_ACL_PKT     4

/**
 * ack pkt arriving error: internal buffer for ack pkt overflow (due to error pkts that arrived)
 */
#define BT_ERR_BUFFER_OVERFLOW        8

/**
 * acl pkt error: service number out of bounds
 */
#define BT_ERR_SERV_NR_OUT_OF_BOUNDS  -1

/**
 * acl pkt error: timeout occured, no pkt with given service number arrived in given time.
 */
#define BT_ERR_ACL_COM_TIMEOUT        -2

/**
 * acl pkt error: playload to large
 */
#define BT_ERR_ACL_PAYLOAD_OVERFLOW   -3

/**
 * Use this packet type to send pkts with a payload up to #BT_MAX_ACL_COM_PAYLOAD.\n
 * DM3 means three slots are used for one pkt. Smaller amount of slots can lead
 * to fragmented pkts
 */
#define HCI_PACKET_TYPE BT_HCI_PACKET_TYPE_DM3

/**
 * for not having fragmented pkts we limit the payload of acl for packet type 
 * DM3 to 117!! because DM3 acl payload limited to 4+117=121 bytes
 */
#define BT_MAX_ACL_COM_PAYLOAD 117

/**
 * defines the length of the acl communication header
 */
#define BT_ACL_COM_HEADER_LEN 4

/**
 * this typedef allocates just enough memory for a acl com pkt with all
 * the header space needed
 */
typedef u_char bt_acl_com_pkt_t[1+HCI_ACL_DATA_PKT_HDR+BT_ACL_COM_HEADER_LEN+BT_MAX_ACL_COM_PAYLOAD];



/** 
 * struct starts with the acl pkt. the con handle holds the handle where the
 * pkt arrived. the state tells about the usage of this buffer entry and
 * the next value points to next entry with same service nr
 */
typedef struct _bt_acl_com_buffer_entry {
    bt_acl_com_pkt_t pkt;
    bt_hci_con_handle_t con_handle;
    u_char state;
    struct _bt_acl_com_buffer_entry *next;
} bt_acl_com_buffer_entry_t;

/**
 * struct to hold for every service (max #MAX_SERVICE_NR) a registered
 * event handle to notify a thread and a list of pointer
 * to packets waiting for this thread. service nr equal to the position
 * in the services array!
 */
typedef struct _bt_acl_com_service {
    HANDLE event;
    bt_acl_com_buffer_entry_t *buffer;    //buffer chain
} bt_acl_com_service_t;

/**
 * holds all the data needed by the acl communication driver
 */
typedef struct _bt_acl_com_stack {   
    struct btstack *bt_stack;        
    bt_acl_com_service_t services[MAX_SERVICE_NR];
    bt_acl_com_buffer_entry_t *buffer;    //array of buffer_size
    u_char buffer_size;
    u_char error_status;
} bt_acl_com_stack_t;


/**
 * to initialize the acl com stack. 
 * 
 * @param bt_stack Pointer to btstack
 * @param nr_buffer Number of receive-buffer for acl com packets
 * @return pointer to bt acl stack 
 */
bt_acl_com_stack_t* bt_acl_com_init(struct btstack *bt_stack, u_char nr_buffer);

/**
 * prints the bt_addr with format: 00:00:00:00:00:00
 * using normal DEBUGT() call 
 * @see \link btn-terminal.h btn-terminal.h\endlink
 */
void bt_acl_com_print_bt_addr(bt_addr_t addr);

/**
 * prints the bt_addr with format: 00:00:00:00:00:00
 * using special DEBUGT_M() call
 * @see \link btn-terminal.h btn-terminal.h\endlink
 */
void bt_acl_com_print_bt_addr_m(bt_addr_t addr);

/**
 * to dump buffer data of length len
 */
void bt_acl_com_dump_data(u_char* data, u_short len);

/**
 * @brief To access the packet this function will return a pointer to the buffer
 * of the packet with service number nr via the argument pointer. 
 *
 * After usage of the packet it is important to call the bt_acl_com_free_packet
 * to avoid acl packet-buffer overflow!\n
 * Use bt_acl_com_get_data_pointer to get to the payload data pointer!
 * \attention But always pass the original bt_hci_pkt_acl pointer to all functions!
 * 
 * @param _bt_acl_com_stack Pointer to the acl communication stack
 * @param service_nr Requested service number (0 - (#MAX_SERVICE_NR - 1))
 * @param p_p_pkt Pass the address of the pointer to a acl com pkt
 * @param con_handle The connection handle where the pkt arrived on
 * @param len length of payload of this pkt
 * @param timeout_ms timeout in milliseconds. Use NUT_WAIT_INFINITE or 0 for no
 * timeout or 1 to check the availability of packets 
 * @return 0 or possible errors: 
 * - #BT_ERR_ACL_COM_TIMEOUT
 * - #BT_ERR_SERV_NR_OUT_OF_BOUNDS
 */
char bt_acl_com_get_packet(bt_acl_com_stack_t* _bt_acl_com_stack, u_char service_nr, bt_acl_com_pkt_t **p_p_pkt, bt_hci_con_handle_t* con_handle, u_short* len, u_long timeout_ms);

/**
 * @brief To free the buffer of a acl communication pkt.
 * 
 * It is important to free the receive buffer again after usage! 
 * 
 * @param _bt_acl_com_stack Pointer to the acl communication stack
 * @param pkt Packet to be freed
 */
long bt_acl_com_free_packet(bt_acl_com_stack_t* _bt_acl_com_stack, bt_acl_com_pkt_t *pkt);

/**
 * @brief Returns the pointer to the start of the acl com data area
 */
u_char* bt_acl_com_get_data_pointer(bt_acl_com_pkt_t*  pkt);

/**
 * @brief Function to send a acl communication packet.
 * 
 * Alloc new bt_acl_com_pkt_t and use bt_acl_com_get_data_pointer to access the
 * acl com data area. Send the packet to the connection with con_handle passing 
 * the pointer to the bt_acl_com_pkt_t the service number and the length of the 
 * acl-com-payload!\n
 * e.g.:\n
\verbatim
bt_acl_com_pkt_t acl_com_pkt;
u_char* data = bt_acl_com_get_data_pointer(&acl_com_pkt);
strcpy(data, "hello world");
bt_acl_com_send_packet(&acl_com_pkt, <con_handle>, <serv_nr>, 12);
\endverbatim
 * @param _bt_acl_com_stack Pointer to the acl communication stack
 * @param pkt ACL communication packet to be sent
 * @param con_handle Handle of the connection for sending
 * @param service_nr service number (0 - (#MAX_SERVICE_NR-1))
 * @param len length of the acl pkt payload
 * @return 0 or possible errors: 
 * - #BT_ERR_BT_MODULE\n
 * - #BT_ERR_NO_CON\n
 * - #BT_ERR_SERV_NR_OUT_OF_BOUNDS\n
 * - #BT_ERR_ACL_PAYLOAD_OVERFLOW
 */
long bt_acl_com_send_packet(bt_acl_com_stack_t* _bt_acl_com_stack, bt_acl_com_pkt_t *pkt, bt_hci_con_handle_t con_handle, u_short service_nr, u_short len);

/**
 * @brief Errors that occure during receiving an acl pkt can be accessed with this
 * function.
 *  
 * After the call the internal error status will be set to 0 again.\n
 * Possible errors:\n
 - #BT_ERR_INVALID_SERV_NR : can only occure if communication error occured\n
 - #BT_ERR_INVALID_HANDLE  : can only occure if communication error occured\n
 - #BT_ERR_FRAGMENTED_ACL_PKT : can occure if communication error occured
                               or pkt bigger than com slots. change pkt type
                               to #HCI_PACKET_TYPE (#BT_HCI_PACKET_TYPE_DM3)\n
 - #BT_ERR_BUFFER_OVERFLOW    : should never occure because module knows how
                               many buffer we have\n
 */
u_char bt_acl_com_get_status(bt_acl_com_stack_t* _bt_acl_com_stack);


#endif

