/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * @file con_power_mgr.h
 *
 * @brief Implementation of a simple power manager
 *        for already established bluetooth connections
 *  
 * \author Mustafa Yuecel <mustafa.yuecel@alumni.ethz.ch>
 * 
 * \date 2006/04/02
 */

#ifndef CON_POWER_MGR_H
#define CON_POWER_MGR_H

#include <sys/types.h>
#include <bt/bt_hci_api.h>

// put 4. slave link to sniff -> no event returned from Zeevo
// Zeevo says that only 1 slave role connection can be putted to sniff

// PARAMS --->
/** maximum links putted to sniff  */
#define CPM_DEFAULT_MAX_SNIFF_LINKS   1
/** put either master or slave links to sniff */
#define CPM_DEFAULT_MASTER_OR_SLAVE   BT_HCI_MY_ROLE_SLAVE

/** periodic time in ms to check the con_table */
#define CPM_DEFAULT_SLEEP_TIME 10000
/** wait time in ms between each link putting in sniff mode */
#define CPM_DEFAULT_WAIT_TIME  1000

/** priority of this thread (see #NutThreadSetPriority) */
#define CPM_DEFAULT_THREAD_PRIORITY   100

/** sniff interval SI (see #bt_hci_sniff_mode) */
#define CPM_DEFAULT_SNIFF_INTERVAL    512
/** sniff attempt SA (see #bt_hci_sniff_mode) */
#define CPM_DEFAULT_SNIFF_ATTEMPT     8
/** sniff timeout ST (see #bt_hci_sniff_mode) */
#define CPM_DEFAULT_SNIFF_TIMEOUT     16
// <--- PARAMS

/**
 * \brief Starts the Connection Power Manager
 * 
 * Puts active links to sniff mode
 */
void cpm_start(void);

/**
 * \brief Pauses the Connection Power Manager
 * 
 * Waits to put active links to sniff mode
 */
void cpm_pause(void);

/**
 * \brief Resumes the Connection Power Manager
 * 
 * Resumes to put active links to sniff mode
 */
void cpm_resume(void);

/**
 * \brief Stops the Connection Power Manager
 * 
 * Returns all sniff links back to active mode
 */
void cpm_stop(void);

/**
 * \brief Initializes the Connection Power Manager
 * 
 * \param stack Pointer to the btstack
 * \param cod only put connections in sniff mode
 *        if remote device has this CoD (0 to disable)
 * 
 * \return 0 if no error 
 */
char cpm_init(struct btstack * bt_stack, u_long cod);

#endif
