/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file logging.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief API for event logging.
 * 
 * \section log_evt Event logging
 * 
 * Logging of events can be done by writing log messages to the log buffer:
 * To log a specific event occuring in your application call
 * #log_add_msg which takes a formatted string and writes it
 * to the log buffer together with the local time #log_add_msg was called.
 * 
 * Log messages consist of a log class, a log level, and the formatted string
 * itself:
 * - the log class specifies the origin of the triggering event and provides
 * a way to distinguish different message types
 * - the log level is used to express the severity of the triggering event
 * 
 * Log classes are assigned by the user before compile time. Note that the
 * btnut system itself uses the event logger, thus log classes above
 * #LOG_SYSTEM_CLASSES have to be chosen.
 * 
 * Currently, four (#LOG_LEVELS) different log levels are supported:
\verbatim
level | define      | description
-------------------------------------------------------------
 1    | LOG_ERROR   | Severe errors that impact the system
 2    | LOG_WARNING | Warnings, indicating non-critical failures
 3    | LOG_INFO    | Informational messages
 4    | LOG_DEBUG   | Debug output intended for developers
\endverbatim
 * 
 * 
 * \section log_stat_masking Static log masking
 * 
 * Static log masks are used to define which log classes and levels are
 * actually compiled and included into the final program code in order that
 * the size of program memory needed for event logging can be controlled.
 * To make use of static log masking, applications typically do not call
 * #log_add_msg directely, but instead use one of the macros below:
\verbatim
#define ERROR(text, ...)
#define WARNING(text, ...)
#define INFO(text, ...)
#define DEBUG(text, ...)
\endverbatim
 * 
 * These macros are available by including the file #debug/log_set.h. For
 * a proper setup, the follwing steps have to be taken:
 * - define the constant LOG_CLASS
 * - define the constant LOG_LEVEL
 * - include #debug/log_set.h
 * 
 * E.g.
\verbatim
#define LOG_CLASS 17
#define LOG_LEVEL 4
#include <log_set.h>
\endverbatim
 * 
 * The macros defined above are then extracted in the following way:
\verbatim
ERROR(text, ...)	-> log_add_msg(LOG_CLASS, LOG_ERROR, text, ...)
WARNING(text, ...)	-> log_add_msg(LOG_CLASS, LOG_WARNING, text, ...)
INFO(text, ...)		-> log_add_msg(LOG_CLASS, LOG_INFO, text, ...)
DEBUG(text, ...)	-> log_add_msg(LOG_CLASS, LOG_DEBUG, text, ...), 
\endverbatim
 *
 * Event masking at compile time is now done by defining LOG_LEVEL
 * appropriately: only messages of levels that are greater or equal than
 * the defined LOG_LEVEL are included in the final program code.
 * 
 * E.g. by setting up logging for log class 17 in the following way:
\verbatim
#define LOG_CLASS 17
#define LOG_LEVEL LOG_WARNING
#include <log_set.h>
\endverbatim
 * only messages of level LOG_ERROR and LOG_WARNING are included in the
 * final program code.
 * 
 * This allows to exclude partial or all (define LOG_LEVEL 0) log information
 * from the final program code for saving program memory.
 * 
 * 
 * \section log_dyn_masking Dynamic log masking
 * 
 * During runtime it is possible to set the logmask, which defines whether
 * a message is written to the log buffer or not. The logmask can be set
 * for each log class individually using #log_set_logmask.
 * 
 * A log mask consists of one bit for each log level and specifies whether
 * the log message shall be written to the log buffer (1) or not (0).
 * 
 * E.g. if you want to prevent debug and info messages of log class 17 being
 * written to the log buffer, call:
\verbatim
log_set_logmask(17, LOG_MASK_ERROR & LOG_MASK_WARNING);
\endverbatim
 * or equivalently:
\verbatim
log_set_logmask(17, 0x03);
\endverbatim
 * 
 * By default, the logmask of all log classes is set to #LOG_FILTER_ALL, i.e. 
 * messages of all levels are written to the log buffer.
 * 
 * 
 * \section log_print Printing
 * 
 * The logged messages can be printed to the terminal by including the
 * file include/terminal/log-cmds.h
 * 
 * \see log-cmds.h
 * 
 * 
 * \section log_setup Setup
 * 
 * A proper setup for using LOG_CLASS 18 for event logging, as well as the
 * terminal to get access to the logger is given below:
\verbatim
#include <debug/logging.h>
#include <terminal/btn-terminal.h>
#include <terminal/log-cmds.h>

// log setup
#define LOG_CLASS 18
#define LOG_LEVEL LOG_DEBUG
#include <log_set.h>

int main(void) {
	...

	// initialize event logger
	log_init();
	
	// initialize terminal
	btn_terminal_init(stdout, "[bt-cmd@btnode]$");
	
	// initialize log cmds
	log_cmds_init(stdout);
	
	...
}
\endverbatim
 */
#ifndef _LOGGING_H_
#define _LOGGING_H_

#include <sys/types.h>

/**
 * \name Log Levels
 * 
 * Definition of the different log levels supported by the event logger.
 */
//@{
/// Number of supported log levels
#define LOG_LEVELS				4

/// Log level: error
#define LOG_ERROR				1

/// Log level: warning
#define LOG_WARNING				2

/// Log level: info
#define LOG_INFO				3

/// Log level: debug
#define LOG_DEBUG				4
//@}

/**
 * \name Log level masks
 * 
 * Definition of the log level masks. Each level mask can be used to
 * filter log messages of a specific class on specific levels.
 * 
 * Setting log masks can be done:
 * - before compile-time to reduce code size (see #log_set.h)
 * - during runtime to log only specific events (see #log_set_logmask,
 * #log_add_msg, #log_write resp.)
 * - during printing to get only relevant messages (see #log_get_messages)
 * - when registering a log event listener to get only messages of interest
 * (see #log_register_event_listener)
 */
//@{
/// To filter errors, i.e. log messages of level #LOG_ERROR
#define LOG_MASK_ERROR		(1 << (LOG_ERROR - 1))

/// To filter warnings, i.e. log messages of level #LOG_WARNING
#define LOG_MASK_WARNING	(1 << (LOG_WARNING - 1))

/// To filter infos, i.e. log messages of level #LOG_INFO
#define LOG_MASK_INFO		(1 << (LOG_INFO - 1))

/// To filter debug statements, i.e. log messages of level #LOG_DEBUG
#define LOG_MASK_DEBUG		(1 << (LOG_DEBUG - 1))

/// To mask no levels
#define LOG_MASK_ALL		0xff

/// To mask all levels
#define LOG_MASK_NONE		0x00
//@}

/**
 * \brief Maximal number of log classes
 */
#define LOG_MAX_CLASSES			32

/**
 * \brief Number of system log classes
 * 
 * System log classes are used internally by the btnut system. Thus, when
 * using the logging facility for applications, make shure that you use
 * a log class > LOG_SYSTEM_CLASS.
 */
#define LOG_SYSTEM_CLASSES		16

/**
 * \brief to print out all log classes
 */
#define LOG_ALL_CLASSES			0

/**
 * \brief to print out all log messages
 */
#define LOG_ALL_MESSAGES		0

/**
 * \brief to turn off log verbosity on all log levels
 */
#define LOG_VERBOSITY_OFF		0

/**
 * \brief Maximal number of optional arguments that can be written to
 * the log buffer
 * 
 * \see #log_add_msg
 */
#define LOG_MAX_NR_ARGS			6


//#define LOG_USE_PROG_MEM

#ifdef LOG_USE_PROG_MEM
#define LOG_PSTR(x) PSTR(x)
#else
#define LOG_PSTR(x) x
#endif

/**
 * \brief Definition ot the log data callback
 * 
 * This callback is called when printing the log (see #log_get_messages)
 */
#define LOG_DATA_CB void (*log_data_cb) (u_char* buf, log_msg_tail* tail, void* cb_arg)

/**
 * \brief Size of the trailer that is attached to each log message
 */
#define LOG_MSG_TAIL_SIZE			(sizeof(log_msg_tail))

/**
 * \brief Definition of a log filter
 * 
 * A log filter is an array of logmasks, defining a logmask for each
 * log class.
 */
typedef u_char log_filter_t[LOG_MAX_CLASSES];

/**
 * \brief Definition of the log message trailer
 * 
 * This trailer is attached to each log message.
 */
typedef struct _log_msg_tail_s {
    u_short id;         ///< Log id
	u_char class;		///< Log class the message corresponds to
	u_char level;		///< Log level of the message
	u_long t_stamp;		///< Time the triggering event occured
	u_short len;		///< Size of the message data
	u_char vcall;
} log_msg_tail;

typedef unsigned int log_msg_args_t[LOG_MAX_NR_ARGS];


/**
 * \brief Initializes event logging
 * 
 * Initializes the log buffer for event logging.
 * 
 * Note: call this function as soon as possible in your systm initialization
 * routine for that events occurring during that phase can be logged
 * properly.
 */
void log_init(void);


/**
 * \brief Adds a message to the log buffer
 * 
 * Basically, this function can be used as you would use a 'printf()'
 * function - there are only two restrictions:
 * - no more than the first #LOG_MSG_NR_ARGS optional arguments passed are
 * actually written to the log buffer
 * - strings are not written to the log buffer: #log_msg_add simply writes
 * the first #LOG_MSG_NR_ARGS optional arguments to the log buffer. Thus,
 * if a string is passed, only the pointer to it will be stored - no data
 * is actually copied to the buffer
 * 
 * Note: applications typically do not call this function directely, but instead
 * use the macros defined in #log_set.h to enable event masking at compile-time.
 * 
 * \param class Log class the message should be added to
 * \param level Log level of the message
 * \param fmt String containing conversion specifications
 * 
 * Note: the function does not check the constraints listed above - this
 * has to be done by the user itself. If you don't feel save with those
 * constraints, use the less-efficient but more save #log_write
 * which has no restrictions compared to a standard printf().
 * 
 * Rem: If the string you would like to log is constant,
 * i.e. resides in either static RAM or in the program memory
 * (#see LOG_USE_PROG_MEM), calling this function is legal because the pointer
 * to the stored string will still be valid when calling e.g. #log_show.
 */
void log_add_msg( u_char class, u_char level, CONST char* fmt, ...);

/**
 * \brief Adds a string message to the log buffer
 * 
 * In contrast to #log_add_msg, this function can be use like a "real"
 * 'printf()' function: the format string \fmt is parsed and the arguments
 * passed are extracted. The resulting output string is then written directly
 * to the log buffer.
 * 
 * Note: it is recommanded to use this function only for logging strings
 * (see remark below).
 * 
 * To make use of static log masking, applications typically do not call this
 * function directely, but instead use one of the following macros:
\verbatim
#define ERROR_STR(text, ...)
#define WARNING_STR(text, ...)
#define INFO_STR(text, ...)
#define DEBUG_STR(text, ...)
\endverbatim
 * Setup and usage are the same as for the macros explained in the header
 * of this documenation.
 * 
 * \param class Log class the string message should be added to
 * \param level Log level of the message
 * \param fmt String containing conversion specifications
 * 
 * Rem: In contrast to #log_add_msg, this function is rather slow due
 * to the parsing and the establishment of the output string. Thus, adding
 * a log message to the log buffer using #log_write may alter your overall
 * system performance considerably.
 */
void log_write(u_char class, u_char level, CONST char* fmt, ...);

/**
 * \brief Empties the log buffer
 */
inline void log_empty(void);

/**
 * \brief Sets the logmask of a log class given
 * 
 * Use this function to enable the logging of events of specific levels at
 * runtime: \p class specifies the log class for which a given logmask
 * \p mask shall be set.
 * 
 * \param class Log class to filter
 * \param mask defines the log levels to consider
 */
void log_set_logmask( u_char class, u_char mask);

/*
 * \brief Returns the current log mask set for a specific class
 * 
 * \param class Log class
 * 
 * \returns
 * - Log level mask of the class passed
 * - -1 if the class passed is invalid
 */
short log_get_logmask(u_char class);

/**
 * \brief Prints the stored log messages using the callback passed.
 * 
 * This function reads messages stored in the log buffer and signals the
 * callback \p LOG_DATA_CB with every message read (see
 * #LOG_DATA_CB). It's the callback that defines how a message is printed.
 * 
 * Various filters can be set for getting relevant messages only:
 * \param max_nr_msgs Maximal number of messages to read. Pass
 * #LOG_ALL_MESSAGES for printing all messages.
 * \param class The log class to print. Pass #LOG_CLASSES_ALL for printing
 * all log classes.
 * \param mask Log levels to print, using a hex-mask for filtering:
 * e.g. pass (LOG_MASK_ERROR & LOG_MASK_WARNING) (i.e. 0x03) to print out error
 * and warning messages only. 
 * 
 * E.g. if you'd like to print out all info and debug messages that have
 * been added to the log class 17, call:
 * 
\verbatim
log_show(17, LOG_MASK_INFO & LOG_MASK_DEBUG, LOG_ALL_MESSAGES);
\endverbatim
 * 
 * or equivalently:
 * 
\verbatim
log_show(17, 0x0c, LOG_ALL_MESSAGES);
\endverbatim
 * \param id_newer Only messages that have a higher id than the given
 * id_newer are sent. Use 0 to get all messages.
 */
void log_get(u_short max_nr_msgs, u_char class, u_char mask, LOG_DATA_CB, void* cb_arg);


void log_get_messages(u_short id_newer, u_short nr_msgs, LOG_DATA_CB, void* cb_arg);


/**
 * \brief Registers an event listener
 * 
 * Registers an event listener that is called each time an event has been
 * triggered, i.e. each time a message had been added to the log
 * buffer, the registered callback \p LOG_DATA_CB is called. Use the
 * log filter \p filter to listen to specific events only
 * (see #log_filter_t, #log_get_messages).
 * 
 * \param class LOG_DATA_CB callback to signal
 * \param cb_arg (optional) callback argument
 * \param filter log filter to use
 */
short log_evt_listener_register(LOG_DATA_CB, void* cb_arg, log_filter_t filter);


void log_evt_listener_clear(void);


void log_evt_listener_pause(void);


void log_evt_listener_resume(void);

#endif
