/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * @file mhop_cl.h
 *
 *\author Kevin Martin <kevmarti@tik.ee.ethz.ch>
 * 
 *\date 2006/03/23
 * 
 * \brief Implementation of a connectionless multi-hop layer,
 * using the L2CAP connectionless data channel.
 * 
 * \par Service Registration
 * 
 * Higher layer services can use the function #mhop_cl_send_pkt to send
 * connection-less multi-hop packets to a specific service
 * on the target device. On the receiving side, the service is identified
 * by a unique "PSM" identifier - a number that is included in the
 * multi-hop packet header.
 * 
 * Receiving connection-less multi-hop data packets is the same as receiving
 * point-to-point data packets: higher layer services
 * have to register a service routine together with a PSM identifier
 * at the protocol/service multiplexor
 * (see #bt_psm_service_register, l2cap_cl.h).
 * 
 * \par Packet Forwarding
 * 
 * Forwarding of a packet to its destination node is done by consulting
 * forwarding tables maintained by each node in the network:
 * if a destination can't be found in the table, the packet is discarded.
 * 
 * Routing is done by broadcasting packets: Each node receiving a broadcast
 * packet stores the source node's address together with the connection
 * handle the packet was received on in the forwarding table.\n
 * Incoming packets having a target address that matches one of the
 * entries in the forwarding table are then forwarded directely over the
 * corresponding connection handle.
 */
#ifndef _MHOP_CL_H_
#define _MHOP_CL_H_

#include <bt/bt_defs.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_psm.h>

/**
 * \name Errors
 */
//@{
/// If amount of data to send is too big
#define MHOP_CL_ERR_PAYLOAD_OVERFLOW			-1
/// Packet send error: no corresponding entry found in the forwarding table
#define MHOP_CL_ERR_FORWARDING					-2
//@}


/**
 * \name Packet Properties
 */
//@{
/// Packet shall be broadcasted
#define MHOP_CL_BROADCAST						0
/// Packet shall be forwarded if possible
#define MHOP_CL_UNICAST							1
/// infinite time to live
#define MHOP_CL_TTL_INFINITE					0
//@}


/**
 * \name Forwarding Table Events
 */
//@{
// Forwarding table change events
/// A new source has been added to the forwarding table
#define MHOP_CL_FWT_SOURCE_ADDED				0
/// A source has been deleted from the forwarding table
#define MHOP_CL_FWT_SOURCE_DELETED				1
/// Forwarding table changed callback
#define MHOP_CL_FWTABLE_CB	\
	void (*fwt_cb) (mhop_cl_fwt_entry* entry, u_char type, void* cb_arg)
//@}


/**
 * \brief Forwarding table entry
 * 
 * When receiving a connectionless l2cap packet containing a multi-hop
 * information trailer whos "source record" flag is set
 * (see typedef #mhop_cl_mhop_ctrl_info), the address of the sender will
 * be stored in the forwarding table.
 * 
 * Essentially, an entry in the forwarding table consists of
 * the address of the source node, and the connection handle over which
 * the source node can be reached.
 */
typedef struct mhop_cl_fwt_entry_s {
    bt_addr_t src_addr;			///< Address of the source node
    bt_hci_con_handle_t handle;	///< Outgoing handle
    u_char seq;                 ///< hold the last seq nr for every src node!
    u_char distance;            ///< distance to the source node
} mhop_cl_fwt_entry;


/**
 * \name Forwarding Table Accessors
 */
//@{

/**
 * Returns the forwarding table entry associated with the passed device address.
 * 
 * Scans the forwarding table for an entry containing the destination
 * address \p dest and returns it.
 * 
 * \param addr Address of the desired destination node
 * 
 * \return
 * - the corresponding forwarding table entry
 * - NULL else
 */
mhop_cl_fwt_entry* mhop_cl_fwt_get_entry(bt_addr_t addr);

/**
 * \brief Updates the forwarding table.
 *
 * \param src_addr Address of the source
 * \param src_handle Handle pkt was received from
 * \param src_seq Sequence number of the source broadcast
 * \param ttl time to live of the source broadcast
 * \param distance Distance to the source
 */
int mhop_cl_fwt_update(bt_addr_t src_addr,
						bt_hci_con_handle_t src_handle,
						u_char src_seq,
						u_char distance);
//@}


/**
 * \name Initialization
 */
//@{
/**
 * \brief Initializes the connection-less multi-hop layer.
 * 
 * \param stack Pointer to the bt stack to use
 * \param psmux Pointer to the protocol/service multiplexor
 * \param mhop_psm PSM the multi-hop layer should use
 * \param buf_size Nr of buffers used for buffering packets that have to
 * be forwarded
 * \param HCI_CON_TABLE_CB_REGISTRATION defines where the mhop layer has
 * to register its callback function that handles changes in the
 * connection table.
 * 
 * \see #bt_psm.h
 */
void mhop_cl_init(struct btstack* bt_stack,
					bt_psm_t* psmux,
					u_short psm,
					u_short buf_size,
					HCI_CON_TABLE_CB_REGISTRATION);
//@}


/**
 * \name Sending multi-hop packets
 */
//@{
/**
 * \brief Sends a multi-hop packet to a service on a remote device
 * 
 * This function sends a data packet to a service on a remote device -
 * if necessary via several relay hops.
 * 
 * Set the broadcast flag \p bc_flag to specify if the packet shall
 * be broadcasted or not:
 * - #BT_MHOP_CL_BROADCAST: the packet is broadcasted, but only the device
 * with BT address \p dest delivers the packet to the application layer.
 * - #BT_MHOP_CL_UNICAST: on intermediate hops, the packet is only
 * forwarded if a corresponding entry can be found in the forwarding table
 * (see explanation on top of this document). Only the device with BT
 * address \p dest delivers the packet to the application layer.
 * 
 * By specifying \p dest = #bt_addr_null, _all_ devices receiving the
 * packet do deliver the packet to the application layer.
 * 
 * \param data Pointer to the data to send
 * \param len Amount of data to send
 * \param dest BT address of the destination
 * \param psm Identifies the service the data shall be delivered to on the
 * receiving side
 * \param bc_flag Determines if the packet should be broadcasted or not
 * \param ttl Maximal time to live for the packet - pass TTL_INFINITE
 * for inifinite time to live
 * 
 * \return 0 or possible errors:
 * - #MHOP_CL_ERR_PAYLOAD_OVERFLOW
 * - #BT_ERR_BT_MODULE
 * - #BT_ERR_NO_CON
 * 
 * Example 1:
\verbatim
u_char data[12];
strcpy(data, "hello world");
mhop_cl_send_pkt(data, 12, <target_addr>, <target_psm>, MHOP_CL_BROADCAST, 5);
\endverbatim
 * sends "hello world" to the device with address 'target_addr'. The packet
 * will be broadcasted, but only the device with BT address <target_addr>
 * will deliver the packet to the service identified with the specified
 * PSM 'target_psm'. The packet will be forwarded at most five times.
 * 
 * Example 2:
\verbatim
u_char data[19];
strcpy(data, "fifa world cup 2006");
mhop_cl_send_pkt(data, 20, <target_addr>, <target_psm>, MHOP_CL_UNICAST, MHOP_CL_TTL_INFINITE);
\endverbatim
 * sends the string "fifa world cup 2006" to the device with address
 * 'target_addr'. The packet is forwarded as long as intermediate hops
 * know how to forward packets to the specified device.
 * 
 * Example 3:
\verbatim
u_char data[12];
strcpy(data, "hello world");
mhop_cl_send_pkt(data, 12, bt_addr_null, <target_psm>, MHOP_CL_BROADCAST, MHOP_CL_TTL_INFINITE);
\endverbatim
 * sends the string "hello world" to _all_ devices in the network.
 * The string "hello world" is passed to the service identified with
 * psm 'target_psm' on _all_ devices in the network.
 */
long mhop_cl_send_pkt(u_char* data,
						u_short len,
						bt_addr_t dest,
						u_short psm,
						u_char bc_flag,
						u_char ttl);

/**
 * \brief broadcasts a source deletion packet
 * 
 * Sends an packet indicating that the device with addr \p addr should
 * be cleared from the forwarding table.
 * 
 * Note: the packet is broadcasted, that is every node in the network
 * deletes the device with addr \p addr from the forwarding table.
 * 
 * \param addr BT address of the device to be deleted
 * \param ttl Maximal time to live for this source deletion packet
 */
long mhop_cl_send_src_del_pkt(bt_addr_t addr, u_char  ttl);
//@}


/**
 * \name Packet accessors
 */
//@{
/**
 * \brief Returns the source address of a received packet.
 */
u_char* mhop_cl_get_source_addr(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the target address of an received packet.
 */ 
u_char* mhop_cl_get_target_addr(struct bt_hci_pkt_acl* pkt);
//@}


/**
 * \brief Returns the bluetooth address of this device.
 * 
 * TODO: move this function to bt_hci.c
 */
u_char* mhop_cl_get_my_addr(void);

/**
 * \brief Returns the maximal amount of data that can be sent within a
 * single multi-hop packet.
 */
u_short mhop_cl_max_payload_size(void);

/**
 * \brief To register the forwarding table change callback.
 * 
 * This function registers the forwarding table change callback function which is
 * called each time a source has been added to / removed from the forwarding table.
 * 
 * \param MHOP_CL_FWTABLE_CB Function-Pointer to the callback
 * \param cb_arg Some callback argument which is passed to the registering
 * layer each time the callback is called
 */
void mhop_cl_fwt_register_cb(MHOP_CL_FWTABLE_CB, void* cb_arg);


void mhop_cl_register_con_table_cb(HCI_CON_TABLE_CB, void* cb_arg);

#endif //_MHOP_CL_H_
